# This file is part of moniplot
#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

""" Moniplot is a Python data visualization library for (satellite)
instrument monitoring.
"""

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass



#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import PurePath

import numpy as np
import xarray as xr


# - local functions --------------------------------
def __get_attrs(dset, field: str) -> dict:
    """Return attributes of the HDF5 dataset.

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the attributes are read
    field : str
       Name of field in compound dataset

    Returns
    -------
    dict with numpy arrays
    """
    _field = None
    if field is not None:
        try:
            _field = {'name': field,
                      'oneof': len(dset.dtype.names),
                      'index': dset.dtype.names.index(field)}
        except Exception as exc:
            raise RuntimeError(
                f'field {field} not found in dataset {dset.name}') from exc
        # print('_field ', _field)

    attrs = {}
    for key in dset.attrs:
        if key in ('CLASS', 'DIMENSION_LIST', 'NAME', 'REFERENCE_LIST',
                   '_Netcdf4Dimid', '_Netcdf4Coordinates'):
            continue

        attr_value = dset.attrs[key]
        # print('# ----- ', key, type(attr_value), attr_value)
        if isinstance(attr_value, np.ndarray):
            if len(attr_value) == 1:
                attr_value = attr_value[0]
                # print('# ----- ', key, type(attr_value), attr_value)
            elif _field is not None:
                if len(attr_value) == _field['oneof']:
                    attr_value = attr_value[_field['index']]
                # elif isinstance(attr_value, np.void):
                #    attr_value = attr_value[0]

        attrs[key] = (attr_value.decode('ascii')
                      if isinstance(attr_value, bytes) else attr_value)

    return attrs


def __get_coords(dset, data_sel: tuple) -> list:
    """Return coordinates of the HDF5 dataset with dimension scales.

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the data is read
    data_sel :  tuple of slice(s)
       A numpy slice generated for example 'numpy.s\_'

    Returns
    -------
    A sequence of tuples [(dims, data), ...]
    """
    coords = []
    if len(dset.dims) == dset.ndim:
        try:
            for ii, dim in enumerate(dset.dims):
                # get name of dimension
                name = PurePath(dim[0].name).name
                if name.startswith('row') or name.startswith('column'):
                    name = name.split(' ')[0]

                # determine coordinate
                buff = None
                if dim[0].size > 0 and not np.all(dim[0][()] == 0):
                    buff = dim[0][()]
                elif name in ('row', 'column'):
                    co_dtype = 'u2' if ((dset.shape[ii]-1) >> 16) == 0 else 'u4'
                    buff = np.arange(dset.shape[ii], dtype=co_dtype)

                if not (buff is None or data_sel is None):
                    buff = buff[data_sel[ii]]

                coords.append((name, buff))
        except RuntimeError:
            coords = []

    return coords


def __set_coords(dset, data_sel: tuple, dims: list) -> list:
    """Set coordinates of the HDF5 dataset.

    Parameters
    ----------
    dset :  h5py.Dataset or numpy.array
       h5py dataset from which the data is read
    data_sel :  tuple of slice(s)
       A numpy slice generated for example 'numpy.s\_'
    dims : list of strings
       Alternative names for the dataset dimensions if not attached to dataset
       Default coordinate names are ['time', ['row', ['column']]]

    Returns
    -------
    A sequence of tuples [(dims, data), ...]
    """
    if dims is None:
        if dset.ndim > 3:
            raise ValueError('not implemented for ndim > 3')

        dims = ['time', 'row', 'column'][-dset.ndim:]

    coords = []
    for ii in range(dset.ndim):
        co_dtype = 'u2' if ((dset.shape[ii]-1) >> 16) == 0 else 'u4'
        buff = np.arange(dset.shape[ii], dtype=co_dtype)
        if data_sel is not None:
            buff = buff[data_sel[ii]]
        coords.append((dims[ii], buff))

    return coords


def __get_data(dset, data_sel: tuple, field: str):
    """Return data of the HDF5 dataset.

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the data is read
    data_sel :  numpy slice
       A numpy slice generated for example 'numpy.s\_'
    field : str
       Name of field in compound dataset or None

    Returns
    -------
    Numpy array

    Notes
    -----
    Read floats always as doubles
    """
    if data_sel is None:
        data_sel = ()

    if np.issubdtype(dset.dtype, np.floating):
        data = dset.astype(float)[data_sel]
        data[data == float.fromhex('0x1.ep+122')] = np.nan
        return data

    if field is None:
        return dset[data_sel]

    data = dset.fields(field)[data_sel]
    if np.issubdtype(data.dtype, np.floating):
        data = data.astype(float)
        data[data == float.fromhex('0x1.ep+122')] = np.nan
    return data


def __check_selection(data_sel: tuple, ndim: int) -> tuple:
    """Check and correct user provided data selection.

    Notes
    -----
    If data_sel is used to select data from a dataset then the number of
    dimensions of data_sel should agree with the HDF5 dataset or one and
    only one Ellipsis has to be used.
    Thus allowed values for data_sel are:
    * [always]: (), np.s\_[:], np.s\_[...]
    * [1-D dataset]: np.s\_[:-1], np.s\_[0]
    * [2-D dataset]: np.s\_[:-1, :], np.s\_[0, :], np.s\_[:-1, 0]
    * [3-D dataset]: np.s\_[:-1, :, 2:4], np.s\_[0, :, :], np.s\_[:-1, 0, 2:4]
    * [Ellipsis] np.s\_[0, ...], np.s\_[..., 4], np.s\_[0, ..., 4]
    """
    if data_sel in (np.s_[:], np.s_[...], np.s_[()]):
        return None

    if np.isscalar(data_sel):
        return (np.s_[data_sel:data_sel+1],)

    buff = ()
    for val in data_sel:
        if val == Ellipsis:
            for _ in range(ndim - len(data_sel) + 1):
                buff += np.index_exp[:]
        elif np.isscalar(val):
            buff += (np.s_[val:val+1],)
        else:
            buff += (val,)

    return buff


# - main function ----------------------------------
def h5_to_xr(h5_dset, data_sel=None, *, dims=None, field=None):
    """Create xarray::DataArray from a HDF5 dataset (with dimension scales).

    Implements a lite interface with the xarray::DataArray, should work for all
    2-D detector images, sequences of detector measurements and trend data.

    Parameters
    ----------
    h5_dset :  h5py.Dataset
       Data, dimensions, coordinates and attributes are read for this dataset
    data_sel :  slice, optional
       A numpy slice generated for example 'numpy.s\_'
    dims :  list of strings, optional
       Alternative names for the dataset dimensions if not attached to dataset
    field : str, optional
       Name of field in compound dataset or None

    Returns
    -------
    xarray.DataArray

    Notes
    -----
    All floating datasets are converted to Python type 'float'

    Dimensions and Coordinates:

    * The functions in this module should work with netCDF4 and HDF5 files.
    * In a HDF5 file the 'coordinates' of a dataset can be defined using \
      dimension scales.
    * In a netCDF4 file this is required: all variables have dimensions, \
      which can have coordinates. But under the hood also netCDF4 uses \
      dimension scales.
    * The xarray DataArray structure will have as dimensions, the names of \
      the dimension scales and as coordinates the names and data of the \
      dimensions scales, except when the data only contains zero's.
    * The default dimensions of an image are 'row' and 'column' with evenly \
      spaced values created with np.arange(len(dim), dtype=uint).

    Examples
    --------
    >>> fid = h5py.File(flname, 'r')        # flname is a HDF5/netCDF4 file
    >>> xdata = h5_to_xr(fid['signal'])
    >>> fid.close()

    Combine Tropomi SWIR data of band7 and band8

    >>> fid = h5py.File(s5p_b7_prod, 'r')   # Tropomi band7 product
    >>> xdata7 = h5_to_xr(fid['signal'])
    >>> fid.close()
    >>> fid = h5py.File(s5p_b8_prod, 'r')   # Tropomi band8 product
    >>> xdata8 = h5_to_xr(fid['signal'])
    >>> fid.close()
    >>> xdata = xr.concat((xdata7, xdata8), dim='spectral_channel')

    Optionally, fix the 'column' dimension

    >>> xdata = xdata.assign_coords(
    >>> ... column=np.arange(xdata.column.size, dtype='u4'))

    """
    # Check data selection
    if data_sel is not None:
        data_sel = __check_selection(data_sel, h5_dset.ndim)

    # Name of this array
    name = PurePath(h5_dset.name).name

    # Values for this array
    data = __get_data(h5_dset, data_sel, field)

    # Coordinates (tick labels) to use for indexing along each dimension
    coords = []
    if dims is None:
        coords = __get_coords(h5_dset, data_sel)
    if not coords:
        coords = __set_coords(h5_dset, data_sel, dims)

    # - check if dimension of dataset and coordinates agree
    if data.ndim < len(coords):
        for ii in reversed(range(len(coords))):
            if np.isscalar(coords[ii][1]):
                del coords[ii]

    # - remove empty coordinates
    dims = []
    co_dict = {}
    for key, val in coords:
        dims.append(key)
        if val is not None:
            co_dict[key] = val

    # Attributes to assign to the array
    attrs = __get_attrs(h5_dset, field)

    return xr.DataArray(data,
                        coords=co_dict, dims=dims, name=name, attrs=attrs)


def data_to_xr(data, *, dims=None, name=None, long_name=None, units=None):
    """Create xarray::DataArray from a dataset.

    Implements a lite interface with the xarray::DataArray, should work for all
    2-D detector images, sequences of detector measurements and trend data.

    Parameters
    ----------
    data :  array-like
       Data to be stored in the xarray
    dims :  list of strings
       Names for the dataset dimensions
    name : str
       A string that names the instance
    units :  str
       Units of the data, default: '1'
    long_name : str
       Long name describing the data, default: empty string

    Returns
    -------
    xarray.DataArray

    Notes
    -----
    All floating datasets are converted to Python type 'float'
    """
    coords = __set_coords(data, None, dims)
    attrs = {}
    attrs['units'] = '1' if units is None else units
    attrs['long_name'] = '' if long_name is None else long_name

    return xr.DataArray(data,
                        coords=coords, name=name, attrs=attrs)



#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from matplotlib.ticker import AutoMinorLocator
import numpy as np

from .fig_legend import blank_legend_key


def fig_draw_qhist(axx, qdata, label: str, density: bool):
    """
    Add a subplot showing pixel-quality data as a histogram

    Parameters
    ----------
    axx :  matplotlib.Axes
    qdata :  numpy.ndarray
       Object holding pixel-quality data and attributes
    label :  str
       Name describing qdata, displayed in the upper left corner
    density :  bool
       See method MONplot::draw_qhist for a description
    """
    qdata[np.isnan(qdata)] = 0.
    # draw histogram
    axx.hist(qdata, bins=10, range=[0, 1], density=density,
             histtype='bar', align='mid', log=True, fill=True,
             edgecolor='#4477AA', facecolor='#77AADD', linewidth=1.5)
    # add label
    legenda = axx.legend([blank_legend_key()],
                         [label], loc='upper left')
    legenda.draw_frame(False)
    # add decoration
    axx.grid(which='major', axis='y', color='#AAAAAA', ls='--')
    axx.set_xlim([0, 1])
    axx.xaxis.set_minor_locator(AutoMinorLocator(2))
    axx.set_ylabel('density' if density else 'count')
    axx.set_ylim([1e-4, 10])
    axx.set_yticks([1e-4, 1e-3, 1e-2, 1e-1, 1])



#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


def get_xylabels(gridspec, data_tuple):
    """
    Define xlabel and ylabel for each subplot panel
    """
    res = ()
    data_iter = iter(data_tuple)
    for _ in range(gridspec.nrows):
        for ix in range(gridspec.ncols):
            data = next(data_iter)
            if isinstance(data, np.ndarray):
                res += (['', ''],)
                continue

            xlabel = data.dims[0]
            ylabel = 'value'
            if 'units' in data.attrs and data.attrs['units'] != '1':
                ylabel += f' [{data.attrs["units"]}]'
            if ix > 0 and ylabel == res[-ix][1]:
                res += ([xlabel, ''],)
            else:
                res += ([xlabel, ylabel],)

    res = np.array(res).reshape((gridspec.nrows, gridspec.ncols, 2))
    for ix in range(gridspec.ncols):
        if np.all(res[:, ix, 0] == res[0, ix, 0]):
            res[:-1, ix, 0] = ''
    return res


def draw_subplot(axx, xarr, xylabels):
    """
    Draw a subplot figure
    """
    if '_plot' in xarr.attrs:
        kwargs = xarr.attrs['_plot']
    else:
        kwargs = {'color': '#4477AA'}

    label = xarr.attrs['long_name'] \
        if 'long_name' in xarr.attrs else None
    xlabel = xarr.dims[0]
    axx.plot(xarr.coords[xlabel], xarr.values, label=label, **kwargs)
    if label is not None:
        _ = axx.legend(fontsize='small', loc='upper right')
    axx.set_xlabel(xylabels[0])
    axx.set_ylabel(xylabels[1])

    if '_title' in xarr.attrs:
        axx.set_title(xarr.attrs['_title'])
    if '_yscale' in xarr.attrs:
        axx.set_yscale(xarr.attrs['_yscale'])
    if '_xlim' in xarr.attrs:
        axx.set_xlim(xarr.attrs['_xlim'])
    if '_ylim' in xarr.attrs:
        axx.set_ylim(xarr.attrs['_ylim'])
    if '_text' in xarr.attrs:
        axx.text(0.05, 0.985, kwargs['text'],
                 transform=axx.transAxes,
                 fontsize='small', verticalalignment='top',
                 bbox=dict(boxstyle='round',
                           facecolor='#FFFFFF',
                           edgecolor='#BBBBBB',
                           alpha=0.5))



#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

import matplotlib.pyplot as plt

from moniplot.lib.fig_info import FIGinfo


def add_fig_box(axx_c, aspect: int, fig_info: FIGinfo) -> None:
    """
    Add a box with meta information for draw_signal and draw_quality

    Parameters
    ----------
    axx_c :  Matplotlib Axes instance of the colorbar
    aspect :  int
    fig_info :  FIGinfo
        instance of moniplot.lib.fig_info to be displayed
    """
    if fig_info is None:
        return

    # put text above colorbar
    if fig_info.location == 'above':
        if aspect <= 2:
            halign = 'center'
            fontsize = 'x-small'
        else:
            halign = 'right'
            fontsize = 'xx-small' if len(fig_info) > 6 else 'x-small'

        axx_c.text(0 if aspect == 2 else 1,
                   1.04 + (aspect-1) * 0.0075,
                   fig_info.as_str(), fontsize=fontsize,
                   transform=axx_c.transAxes,
                   multialignment='left',
                   verticalalignment='bottom',
                   horizontalalignment=halign,
                   bbox={'facecolor': 'white', 'pad': 4})
        return

    if fig_info.location == 'below':
        fontsize = 'xx-small' if aspect in (3, 4) else 'x-small'
        axx_c.text(0.125 + (aspect-1) * 0.2,
                   -0.03 - (aspect-1) * 0.005,
                   fig_info.as_str(), fontsize=fontsize,
                   transform=axx_c.transAxes,
                   multialignment='left',
                   verticalalignment='top',
                   horizontalalignment='left',
                   bbox={'facecolor': 'white', 'pad': 4})


# -------------------------
def draw_figure(aspect: int, side_panels='one') -> None:
    """
    Show figure with given aspect ratio
    """
    attrs = {1: {'figsize': (10, 8),
                 'w_ratios': (1., 7., 0.5, 1.5),
                 'h_ratios': (7., 1.)},                  # 7 x 7
             2: {'figsize': (13, 6.25),
                 'w_ratios': (1., 10., 0.5, 1.5),
                 'h_ratios': (5., 1.)},                  # 10 x 5
             3: {'figsize': (15, 5.375),
                 'w_ratios': (1., 12., 0.5, 1.5),
                 'h_ratios': (4., 1.)},                  # 12 x 4
             4: {'figsize': (17, 5.125),
                 'w_ratios': (1., 14., 0.5, 1.5),
                 'h_ratios': (3.5, 1.)}}.get(aspect)     # 14 x 3.5

    fig = plt.figure(figsize=attrs['figsize'])
    fig.suptitle('test of matplotlib gridspec', fontsize='x-large',
                 position=(0.5, 1 - 0.4 / fig.get_figheight()))
    #
    # Define a grid layout to place subplots within the figure.
    #
    # Parameters:
    #    nrows, ncols :  int
    #        The number of rows and columns of the grid.
    #    left, right, top, bottom :  float, optional
    #        Extent of the subplots as a fraction of figure width or height.
    #        Left cannot be larger than right, and bottom cannot be larger
    #        than top. If not given, the values will be inferred from a figure
    #        or rcParams at draw time. See also GridSpec.get_subplot_params.
    #    wspace :  float, optional
    #        The amount of width reserved for space between subplots, expressed
    #        as a fraction of the average axis width. If not given, the values
    #        will be inferred from a figure or rcParams when necessary.
    #        See also GridSpec.get_subplot_params.
    #    hspace :  float, optional
    #        The amount of height reserved for space between subplots,
    #        expressed as a fraction of the average axis height. If not given,
    #        the values will be inferred from a figure or rcParams when
    #        necessary. See also GridSpec.get_subplot_params.
    #    width_ratios :  array-like of length ncols, optional
    #        Defines the relative widths of the columns. Each column gets a
    #        relative width of width_ratios[i] / sum(width_ratios). If not
    #        given, all columns will have the same width.
    #    height_ratios :  array-like of length nrows, optional
    #        Defines the relative heights of the rows. Each row gets a
    #        relative height of height_ratios[i] / sum(height_ratios).
    #        If not given, all rows will have the same height.
    gspec = fig.add_gridspec(2, 4,
                             wspace=0.1 / max(2, aspect-1), hspace=0.05,
                             width_ratios=attrs['w_ratios'],
                             height_ratios=attrs['h_ratios'],
                             left=.135 + .005 * (aspect-1),
                             right=.9 - .005 * (aspect-1),
                             bottom=.115 + .01 * (aspect-1),
                             top=.865 - .025 * (aspect-1))
    # else:
    #    gspec = fig.add_gridspec(2, 4, wspace=0.05 / aspect, hspace=0.025,
    #                             width_ratios=(1., 4. * aspect, .3, .7),
    #                             height_ratios=(4, 1),
    #                             bottom=.1 if aspect == 2 else .125,
    #                             top=.85 if aspect == 2 else 0.825)

    axx = fig.add_subplot(gspec[0, 1])
    pcm = axx.pcolormesh(np.random.randn(30, aspect * 30),
                         vmin=-2, vmax=2)
    # image panel
    axx.set_title(f'aspect={aspect}')
    axx.grid(True)

    # side_panels
    if side_panels != 'none':
        for xtl in axx.get_xticklabels():
            xtl.set_visible(False)
        for ytl in axx.get_yticklabels():
            ytl.set_visible(False)
        axx_px = fig.add_subplot(gspec[1, 1], sharex=axx)
        axx_px.set_xlabel('column')
        axx_px.grid(True)
        axx_py = fig.add_subplot(gspec[0, 0], sharey=axx)
        axx_py.set_ylabel('row')
        axx_py.grid(True)
    else:
        axx.set_xlabel('column')
        axx.set_ylabel('row')

    # add colorbar
    axx_c = fig.add_subplot(gspec[0, 2])
    _ = plt.colorbar(pcm, cax=axx_c, label='value')

    fig_info = FIGinfo()
    for ii in range(5):
        fig_info.add(f'text line {ii+1}', 'blah blah blah')
    add_fig_box(axx_c, aspect, fig_info)

    plt.show()


def main():
    """
    Main function
    """
    draw_figure(1)
    draw_figure(2)
    draw_figure(3)
    draw_figure(4)


# --------------------------------------------------
if __name__ == '__main__':
    main()



#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numbers import Integral
import numpy as np

from matplotlib.ticker import AutoMinorLocator, MultipleLocator

from ..tol_colors import tol_cset

from .fig_legend import blank_legend_key


# - local functions --------------------------------
def set_labels_colors(xarr) -> tuple:
    """
    Determine name and units of housekeeping data and line and fill color

    Returns
    -------
    tuple
       sub_title, ylabel, line color and fill color
    """
    cset = tol_cset('bright')

    hk_unit = xarr.attrs['units']
    if isinstance(hk_unit, bytes):
        hk_unit = hk_unit.decode()

    hk_title = xarr.attrs['long_name']
    if isinstance(hk_title, bytes):
        hk_title = hk_title.decode()

    if hk_unit == 'K':
        if (ii := hk_title.find(' temperature')) > 0:
            hk_title = hk_title[:ii]
        hk_label = f'temperature [{hk_unit}]'
        lcolor = cset.blue
        fcolor = '#BBCCEE'
    elif hk_unit in ('A', 'mA'):
        if (ii := hk_title.find(' current')) > 0:
            hk_title = hk_title[:ii]
        hk_label = f'current [{hk_unit}]'
        lcolor = cset.green
        fcolor = '#CCDDAA'
    elif hk_unit == '%':
        if (ii := hk_title.find(' duty')) > 0:
            hk_title = hk_title[:ii]
        hk_label = f'duty cycle [{hk_unit}]'
        lcolor = cset.red
        fcolor = '#FFCCCC'
    else:
        hk_label = f'value [{hk_unit}]'
        lcolor = cset.purple
        fcolor = '#EEBBDD'

    # overwrite ylabel
    if '_ylabel' in xarr.attrs:
        hk_label = xarr.attrs['_ylabel']

    return hk_title, hk_label, lcolor, fcolor


def adjust_ylim(avg, err1, err2, vperc: list, vrange_last_orbits: int):
    """
    Return the limits of the Y-coordinate
    """
    if err1 is not None and err2 is not None:
        indx = np.isfinite(err1) & np.isfinite(err2)
        if np.all(~indx):
            ylim = [0., 0.]
        elif np.sum(indx) > vrange_last_orbits > 0:
            ni = vrange_last_orbits
            ylim = [min(err1[indx][0:ni].min(), err1[indx][-ni:].min()),
                    max(err2[indx][0:ni].max(), err2[indx][-ni:].max())]
        elif isinstance(vperc, list) and len(vperc) == 2:
            ylim = [np.percentile(err1[indx], vperc[0]),
                    np.percentile(err2[indx], vperc[1])]
        else:
            ylim = [err1[indx].min(), err2[indx].max()]
        factor = 10
    else:
        indx = np.isfinite(avg)
        if np.all(~indx):
            ylim = [0., 0.]
        elif np.sum(indx) > vrange_last_orbits > 0:
            ni = vrange_last_orbits
            ylim = [min(avg[indx][0:ni].min(), avg[indx][-ni:].min()),
                    max(avg[indx][0:ni].max(), avg[indx][-ni:].max())]
        elif isinstance(vperc, list) and len(vperc) == 2:
            ylim = np.percentile(avg[indx], vperc)
        else:
            ylim = [avg[indx].min(), avg[indx].max()]
        factor = 5

    if ylim[0] == ylim[1]:
        delta = 0.01 if ylim[0] == 0 else ylim[0] / 20
    else:
        delta = (ylim[1] - ylim[0]) / factor

    return (ylim[0] - delta, ylim[1] + delta)


def adjust_units(zunit: str):
    """
    Adjust units: electron to 'e' and Volt to 'V'

    Parameters
    ----------
    zunit :  str
       Units of the image data

    Returns
    -------
    str
    """
    if zunit is None or zunit == '1':
        return '1'

    if zunit.find('electron') >= 0:
        zunit = zunit.replace('electron', 'e')
    if zunit.find('Volt') >= 0:
        zunit = zunit.replace('Volt', 'V')
    if zunit.find('.s-1') >= 0:
        zunit = zunit.replace('.s-1', ' s$^{-1}$')

    return zunit


def get_gap_list(xdata: np.ndarray) -> list:
    """
    Identify data gaps
    """
    if not issubclass(xdata.dtype.type, Integral):
        return []

    uvals, counts = np.unique(np.diff(xdata), return_counts=True)
    if counts.size > 1 and counts.max() / xdata.size > 0.5:
        xstep = uvals[counts.argmax()]
        return (np.diff(xdata) > xstep).nonzero()[0].tolist()

    return []


# - main functions ---------------------------------
def add_subplot(axx, xarr) -> None:
    """
    Add a subplot for measurement data

    Parameters
    ----------
    axx :  matplotlib.Axes
    xarr :  xarray.DataArray
       Object holding measurement data and attributes
    """
    cset = tol_cset('bright')
    ylabel = xarr.attrs['long_name']
    if 'units' in xarr.attrs and xarr.attrs['units'] != '1':
        ylabel += f' [{adjust_units(xarr.attrs["units"])}]'
    lcolor = xarr.attrs['_color'] if '_color' in xarr.attrs else cset.blue
    fcolor = '#BBCCEE'

    # define xdata and determine gap_list (always atleast one element!)
    if 'orbit' in xarr.coords:
        xdata = xarr.coords['orbit'].values
        isel = np.s_[:]
    else:
        xdata = xarr.coords['time'].values
        isel = np.s_[0, :]
    gap_list = get_gap_list(xdata)
    gap_list.append(xdata.size - 1)

    # define avg, err1, err2
    # check if xarr.values is a structured array:
    #    xarr.values.dtype.names is None
    # check if xarr contains quality data
    # check if err1 and err2 are present
    if xarr.values.dtype.names is None:
        avg = xarr.values[isel]
        err1 = err2 = None
    else:
        avg = xarr.values['mean'][isel]
        err1 = xarr.values['err1'][isel]
        err2 = xarr.values['err2'][isel]

    ii = 0
    for jj in gap_list:
        isel = np.s_[ii:jj+1]
        if err1 is not None:
            axx.fill_between(xdata[isel], err1[isel], err2[isel],
                             step='post', linewidth=0, facecolor=fcolor)
            axx.step(np.append(xdata[isel], xdata[jj]),
                     np.append(avg[isel], avg[jj]), where='post',
                     linewidth=1.5, color=lcolor)
        else:
            axx.plot(xdata[isel], avg[isel], linewidth=1.5, color=lcolor)
        if 'legend' in xarr.attrs:
            legenda = axx.legend([blank_legend_key()],
                                 [xarr.attrs['legend']], loc='upper left')
            legenda.draw_frame(False)
        ii = jj + 1

    # adjust data X-coordinate
    if 'time' in xarr.coords:
        axx.xaxis.set_major_locator(MultipleLocator(3))
        axx.xaxis.set_minor_locator(MultipleLocator(1))
    else:
        axx.xaxis.set_minor_locator(AutoMinorLocator())
    axx.set_xlim([xdata[0], xdata[-1]])

    # adjust data X-coordinate
    axx.locator_params(axis='y', nbins=5)
    if 'orbit' in xarr.coords:
        axx.set_ylim(adjust_ylim(avg, err1, err2, None, -1))

    axx.set_ylabel(ylabel)
    axx.grid(True)


def add_hk_subplot(axx, xarr, vperc=None, vrange_last_orbits=-1) -> None:
    """
    Add a subplot for housekeeping data

    Parameters
    ----------
    axx :  matplotlib.Axes
    xarr :  xarray.DataArray
       Object holding housekeeping data data and attributes.
       Dimension must be 'orbit', 'hours' or 'time'.
    vperc :  list, optional
       Reject outliers before determining vrange
       (neglected when vrange_last_orbits is used)
    vrange_last_orbits :  int
        Use the last N orbits to determine vrange (orbit coordinate only)
    """
    hk_title, hk_label, lcolor, fcolor = set_labels_colors(xarr)

    # define xdata and determine gap_list (always one element!)
    if 'orbit' in xarr.coords:
        xdata = xarr.coords['orbit'].values
        isel = np.s_[:]
        gap_list = get_gap_list(xdata)
    elif 'hours' in xarr.coords:
        xdata = xarr.coords['hours'].values
        isel = np.s_[0, :]
        gap_list = get_gap_list(np.round(3600 * xdata).astype(int))
    else:
        xdata = xarr.coords['time'].values
        isel = np.s_[0, :]
        gap_list = get_gap_list(xdata)
    gap_list.append(xdata.size - 1)

    # define avg, err1, err2
    avg = xarr.values['mean'][isel]
    err1 = xarr.values['err1'][isel]
    err2 = xarr.values['err2'][isel]

    # plot data
    ii = 0
    for jj in gap_list:
        isel = np.s_[ii:jj+1]
        axx.fill_between(xdata[isel], err1[isel], err2[isel],
                         step='post', linewidth=0, facecolor=fcolor)
        axx.plot(xdata[isel], avg[isel], linewidth=1.5, color=lcolor)
        ii = jj + 1

    # adjust data X-coordinate
    if 'hours' in xarr.coords:
        axx.xaxis.set_major_locator(MultipleLocator(3))
        axx.xaxis.set_minor_locator(MultipleLocator(1))
    else:
        axx.xaxis.set_minor_locator(AutoMinorLocator())
    axx.set_xlim([xdata[0], xdata[-1]])

    # adjust data Y-coordinate
    axx.locator_params(axis='y', nbins=4)
    if 'orbit' in xarr.coords:
        axx.set_ylim(adjust_ylim(avg, err1, err2, vperc, vrange_last_orbits))
    axx.set_ylabel(hk_label)
    axx.grid(True)

    # add hk_title inside current subplots
    legenda = axx.legend([blank_legend_key()],
                         [hk_title], loc='upper left')
    legenda.draw_frame(False)



#
# https://github.com/rmvanhees/moniplot.git
#
# Copyright (c) 2022 SRON - Netherlands Institute for Space Research
#
# License:  GPLv3
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

try:
    from cartopy import crs as ccrs
except ModuleNotFoundError:
    FOUND_CARTOPY = False
else:
    FOUND_CARTOPY = True
from matplotlib.patches import Polygon
import matplotlib.pyplot as plt

from ..tol_colors import tol_cset


# - main functions ---------------------------------
if FOUND_CARTOPY:
    def fig_draw_tracks(axx, lons, lats, icids, saa_region) -> None:
        """
        Draw satellite tracks projected on the Earth surface.
        """
        # define colors
        cset = tol_cset('bright')

        # draw coastlines and gridlines
        axx.set_global()
        axx.coastlines(resolution='110m')
        axx.gridlines()

        # draw satellite position(s)
        icolor = 0
        for val in np.unique(icids):
            mask = icids == val
            # pylint: disable=abstract-class-instantiated
            plt.plot(lons[mask], lats[mask], linestyle='',
                     marker='s', markersize=2,
                     color=cset[icolor % 6], label=f'ICID: {val}',
                     transform=ccrs.PlateCarree())
            icolor += 1
        axx.legend(loc='lower left')

        # draw SAA region
        if saa_region is not None:
            # pylint: disable=abstract-class-instantiated
            saa_poly = Polygon(xy=saa_region, closed=True, alpha=1.0,
                               facecolor=cset.grey,
                               transform=ccrs.PlateCarree())
            axx.add_patch(saa_poly)
